
BSR <- function(train.data, test.data, outdir, response.var, id.var, vars0, groups=NULL,
         interactions=NULL,  op=NULL) {

  # op             List of options
  #  train.op      Options for training
  #  resample.op   Options for resampling
  #  out.string    String to be appended to output files
  #  cut.model
  #  n.burn
  
  if (!is.data.frame(train.data)) stop("train.data must be a data frame")
  if (!is.data.frame(test.data)) stop("test.data must be a data frame")
  if (ncol(train.data) != ncol(test.data)) stop("train.data and test.data have a different number of columns")
  if (any(colnames(train.data) != colnames(test.data))) {
    stop("train.data and test.data do not contain the same variable names")
  }
  vars0  <- unique(vars0)
  nvars0 <- length(vars0)
  if (!nvars0) stop("vars0 must have at least one variable (intercept)")

  # Remove missing values and check for intercept
  temp  <- samc.modifyData(train.data, response.var, id.var, vars0)
  if (!is.null(temp[["data", exact=TRUE]])) train.data <- temp$data
  temp  <- samc.modifyData(test.data, response.var, id.var, vars0)
  if (!is.null(temp[["data", exact=TRUE]])) test.data <- temp$data
  rm(temp)
  gc()


  if (is.null(outdir)) outdir <- getwd()
  out <- checkForSep(outdir)

  op          <- default.list(op, c("out.string"), list(""))
  out.str     <- op$out.string

  train.op <- op[["train.op", exact=TRUE]]
  if (is.null(train.op)) train.op <- list()
  resample.op <- op[["resample.op", exact=TRUE]]
  if (is.null(resample.op)) resample.op <- list()
  prediction.op <- op[["prediction.op", exact=TRUE]]
  if (is.null(prediction.op)) prediction.op <- list()

  # Set some default option for training if not specified
  train.op <- default.list(train.op, c("maxvar", "maxbeta", "gamma"), list(25, 10.0, 0.5))

  # Train
  train.op$out.string <- paste("TRAIN", out.str, sep="")
  train.op$check.data <- 0
  cat("Calling BSR.samc\n")
  ret.train <- BSR.samc(train.data, outdir, response.var, id.var, vars0, groups=groups, 
                interactions=interactions, op=train.op)

  # Resample
  resample.op$out.string <- paste("RESAMPLE", out.str, sep="")
  resample.op$check.data <- 0
  resample.op$in_models  <- ret.train$out_models
  resample.op$in_hist    <- ret.train$out_hist
  resample.op$maxvar     <- train.op$maxvar
  resample.op$maxbeta    <- train.op$maxbeta
  resample.op$gamma      <- train.op$gamma
  cat("Calling BSR.sample.model\n")
  ret.resample <- BSR.sample.model(train.data, outdir, response.var, id.var, vars0, groups=groups,
                   interactions=interactions,  op=resample.op)

  # Prediction
  model.file <- ret.resample$out2
  gof.file   <- ret.resample$out1
  best.file  <- ret.resample$out_best
  prediction.op$out.string <- paste("PREDICTION", out.str, sep="")
  cat("Calling BSR.prediction\n")
  ret <- BSR.prediction(train.data, test.data, outdir, response.var, id.var, 
                  model.file, gof.file, best.file, op=prediction.op)
   
  
  delete <- train.op[["delete", exact=TRUE]]
  if (is.null(delete)) delete <- 0
  if (delete) {
    ff <- unlist(ret.train)
    for (f in ff) {
      if (!is.null(f)) { 
        if (file.exists(f)) file.remove(f)
      }
    } 
  }
  delete <- resample.op[["delete", exact=TRUE]]
  if (is.null(delete)) delete <- 0
  if (delete) {
    ff <- unlist(ret.resample)
    for (f in ff) {
      if (!is.null(f)) { 
        if (file.exists(f)) file.remove(f)
      }
    } 
  }

  ret

} # END: samc

# Function to check the input data
samc.modifyData <- function(data, response.var, id.var, vars0) {

  dim0    <- dim(data)
  nvars0  <- length(vars0)
  id.flag <- !is.null(id.var)

  # Check for constant variables and remove missing values
  cnames <- colnames(data)
  if (id.flag) {
    temp   <- !(cnames %in% id.var)
    cnames <- cnames[temp]
  }
  nr     <- nrow(data)
  temp   <- rep(TRUE, nr)
  for (v in cnames) {
    vec <- data[, v]
    if (!is.numeric(vec)) {
      str <- paste("Variable ", v, " is not numeric", sep="")
      stop(str)
    }
    temp <- temp & (!is.na(vec))
  }
  if (!all(temp)) data <- removeOrKeepRows(data, temp)
  if (!nrow(data)) stop("All rows in data have been removed due to missing values")

  # Check the response
  if (!all(data[, response.var] %in% 0:1)) stop("response.var must be coded as 0 and 1")

  if (id.flag) {
    temp <- removeOrKeepCols(data, id.var, which=-1)
  } else {
    temp <- data
  }
  temp <- checkForConstantVar(temp, msg=0, removeVars=0)$remove
  n    <- length(temp)
  if (n) {
    consVar <- temp
    if (response.var %in% consVar) stop("response.var is constant")
     
    if (n > 1) {
      print(consVar)
      print("The above variables are constant. There should be no more than 1 constant variable (intercept)")
      warning("Constant variables in data frame")
    }

    # Check for intercept
    found  <- 0
    intVar <- NULL
    for (v in consVar) {
      if (all(data[, v] %in% 1)) {
        found  <- 1
        intVar <- v
        break
      }
    }
    if (!found) {
      warning("An intercept column was not found in the data")
    } else {
      if (!(intVar %in% vars0)) warning("vars0 does not contain the intercept variable")
    }
  } else {
    warning("An intercept column was not found in the data")
  }

  #nvars0 <- length(vars0)
  #if (!nvars0) {
  #  # Create the intercept column
  #  vars0 <- "MYINTERCEPT.xcv5342" 
  #  data[, vars0] <- 1
  #}

  # See if data was modified
  if (all(dim(data) == dim0)) data <- NULL
 
  list(data=data, vars0=vars0)

} # END: samc.modifyData

# Function for prediction
BSR.prediction <- function(train.data, test.data, outdir, response.var, id.var, model.file, 
               gof.file, best.file, op=NULL) {

  op <- default.list(op, c("n.burn", "cut.model", "out.string"), 
         list(100, 1000, ""))

  if (is.null(outdir)) outdir <- getwd()
  out <- checkForSep(outdir)

  n.burn    <- op$n.burn
  cut.model <- op$cut.model
  id.str    <- op$out.string
 
  nc      <- length(scan(gof.file, what="character", sep="", nlines=1, quiet=TRUE))
  if (!nc) stop("gof.file is empty")
  gof.mat <- matrix(scan(gof.file, what="character", sep="", quiet=TRUE), byrow=TRUE, ncol=nc)
  n.model <- dim(gof.mat)[1]
  z.vec   <- as.numeric(gof.mat[,6])
  rm(gof.mat)
  gc()
 
  if (n.burn > n.model) n.burn <- n.model
  id.select <- rep(FALSE, n.model)
  temp <- (z.vec < cut.model)
  temp[is.na(temp)]   <- FALSE
  id.select[temp]     <- TRUE
  id.select[1:n.burn] <- FALSE

  if (!any(id.select)) warning("No rows selected. Modify n.burn and/or cut.model.")

  nr.train       <- nrow(train.data)    
  nr             <- nrow(test.data)
  pred.train.ave <- rep(0, nrow(train.data))
  pred.test.ave  <- rep(0, nr)

  # Read in the out2 data from resample. First is the number of dimensions,
  # then the column ids, then the betas in alternating rows
  fid <- file(model.file, "r")
  n   <- 0

  for (i in 1:n.model)
  {
    dim.all <- scan(fid, what=double(0), sep=" ", nlines=1, quiet=TRUE)
    cols    <- scan(fid, what=double(0), sep=" ", nlines=1, quiet=TRUE)
    beta    <- scan(fid, what=double(0), sep=" ", nlines=1, quiet=TRUE)
    if (!length(beta)) break
    if (id.select[i] == 1)
    {
        dim <- dim.all
        # Compute the probs
        dim(beta) <- c(dim, 1)
        mat       <- as.matrix(test.data[, cols]) ## check if it work for one subject
        dim(mat)  <- c(nr, dim)
        t.test    <- exp(mat %*% beta)
        pred.vec  <- t.test /(1+t.test) 
 
        mat.train <- as.matrix(train.data[, cols])
        dim(mat.train) <- c(nr.train, dim)
        t.1 <- exp(mat.train %*% beta)

        pred.train.ave <- pred.train.ave + (t.1/(1+t.1))
        pred.test.ave  <- pred.test.ave + pred.vec
        n <- n + 1
    }
  }
  close(fid) 

  pred.train.ave <- pred.train.ave/n
  pred.test.ave  <- pred.test.ave/n

  # Compute errors
  error.train.ave <- abs(pred.train.ave - train.data[, response.var])
  error.test.ave  <- abs(pred.test.ave - test.data[, response.var])

  # Compute the errors based on the best model.
  fid   <- file(best.file, "r")
  nbeta <- scan(fid, what=double(0), sep=" ", nlines=1, quiet=TRUE)
  cols  <- scan(fid, what=double(0), sep=" ", nlines=1, quiet=TRUE)
  beta  <- scan(fid, what=double(0), sep=" ", nlines=1, quiet=TRUE)
  close(fid)
  dim(beta) <- c(nbeta, 1)
  mat       <- as.matrix(test.data[, cols]) ## check if it work for one subject
  dim(mat)  <- c(nr, nbeta)
  t.test    <- exp(mat %*% beta)
  pred.vec  <- t.test/(1+t.test) 

  mat.train <- as.matrix(train.data[, cols])
  dim(mat.train) <- c(nr.train, nbeta)  
  t.1  <- exp(mat.train %*% beta)
  pred <- t.1/(1+t.1)

  # Assume first column is the response
  #error.train <- abs(pred - train.data[, response.var])
  #error.test  <- abs(pred.vec - test.data[, response.var])

  # Last column are the ids
  train <- cbind(pred.train.ave, pred, train.data[, response.var], train.data[, id.var])
  colnames(train) <- c("PRED.AVG", "PRED.BEST", response.var, id.var)

  test <- cbind(pred.test.ave, pred.vec, test.data[, response.var], test.data[, id.var])
  colnames(test) <- colnames(train)

  temp1 <- paste(outdir, id.str, ".TRAIN.txt.xls", sep="")
  writeTable(train, temp1)

  temp2 <- paste(outdir, id.str, ".TEST.txt.xls", sep="")
  writeTable(test, temp2)

  ret <- list(out.train=temp1, out.test=temp2)
  
  ret

} # END: BSR.prediction


######################################################################
# Functions to be used for an R package 
######################################################################

# Function to add an intercept column (column 1)
addInt <- function(x, name="Intercept") {

  x1 <- data.frame(rep(1, times=nrow(x)))
  colnames(x1) <- name
  return(cbind(x1, x))

} # END: addInt

# Function to get column numbers from variables
colNum <- function(x, vars) {

  ret    <- match(vars, colnames(x))
  temp   <- is.na(ret)
  if (any(temp)) {
    print("The following variables were not found in the data:")
    print(vars[temp])
    stop()
  }
  
  ret

} # END: colNum

# Function to create the interactions file
get_inter <- function(interactions, groups.name, groups.num, inter.file) {

  n     <- length(interactions)
  x     <- character(n)
  index <- 1:length(groups.name) 
  used  <- character(n)

  for (i in 1:n) {
    vec  <- interactions[[i]]
    if (length(vec) < 2) stop(paste("interactions vector ", i, " is incorrect", sep=""))
    if (any(duplicated(vec))) stop("interactions have duplicated groups")
    igrp    <- vec[1]
    if (igrp %in% used) stop(paste("interactions group ", igrp, " is defined more than once", sep="")) 
    used[i] <- igrp 
    ids     <- match(vec, groups.name)
    grp     <- groups.num[ids]
    x[i]    <- paste(grp, collapse=",", sep="")
  }
  write(x, file=inter.file, ncolumns=1)

} # END: get_inter

# Main function for samc or resample
mainfunc <- function(data, outdir, response.var, id.var, vars0, func, 
              groups=NULL, interactions=NULL, op=NULL) {

  if (!is.data.frame(data)) stop("data must be a data frame")
  if (ncol(data) < 2) stop("data must contain at least 2 columns")
  vars0  <- unique(vars0)
  nvars0 <- length(vars0)
  if (!nvars0) stop("vars0 must have at least one variable (intercept)")
  dim0 <- dim(data)
  
  group.flag <- length(groups)
  inter.flag <- length(interactions)
  id.flag    <- !is.null(id.var)
  if ((inter.flag) && (!group.flag)) stop("groups must be specified for the interactions argument")
  if (group.flag) {
    if (!is.list(groups)) stop("groups must be a list")
  }
  if (inter.flag) {
    if (!is.list(interactions)) stop("interactions must be a list")
  }
 

  op <- default.list(op, c("stepscale", "total_iteration", "warm", "savestep", 
                           "scale", "rho", "temperature",
                           "maxvar", "maxbeta", "seed", "gamma", "model_maxn",
                           "delete", "check.data"),
                            list(1000, 500000, 10000, 1, 
                                 1.0, 1.0, 1.0,
                                 25, 10.0, NULL, 0.5, 100000,
                                 0, 1))
  if (func == "samc") {
    if (is.null(op[["lowE", exact=TRUE]])) stop("The option lowE must be specified")
    if (is.null(op[["maxE", exact=TRUE]])) stop("The option maxE must be specified")
  } else {
    # Set to any value since they will be ignored
    op$lowE <- 0
    op$maxE <- 0
  }
  if (is.null(op[["Model_maxn", exact=TRUE]])) op$Model_maxn <- op$model_maxn
  op$model_maxn <- NULL

  if (is.null(op[["out.string", exact=TRUE]])) {
    if (func == "samc") {
      op$out.string <- "TRAIN"
    } else {
      op$out.string <- "RESAMPLE"
    }
  }

  # Check input files
  temp <- op[["in_models", exact=TRUE]]
  if (!is.null(temp)) {
    if (!file.exists(temp)) stop(paste("ERROR: the file ", temp, " was not found", sep=""))
  }
  temp <- op[["in_hist", exact=TRUE]]
  if (!is.null(temp)) {
    if (!file.exists(temp)) stop(paste("ERROR: the file ", temp, " was not found", sep=""))
  }

  delete        <- op$delete
  op$delete     <- NULL
  str           <- op$out.string
  op$out.string <- NULL
  groups.col    <- NULL
  inter.col     <- NULL
  
  # Create output file names
  if (is.null(outdir)) outdir <- getwd()
  outdir        <- checkForSep(outdir)
  op$out_log    <- paste(outdir, str, ".log", sep="")
  op$out_hist   <- paste(outdir, str, ".hist", sep="")
  op$out1       <- paste(outdir, str, ".out1", sep="")
  op$out2       <- paste(outdir, str, ".out2", sep="")
  op$out_best   <- paste(outdir, str, ".best", sep="")
  op$out_models <- paste(outdir, str, ".models", sep="")
  outnames      <- c("out_hist", "out1", "out2", "out_best", "out_models", "out_log") 

  data.file     <- paste(outdir, str, ".data", sep="")
  group.file    <- paste(outdir, str, ".groups", sep="")
  parm.file     <- paste(outdir, str, ".parms", sep="")
  inter.file    <- paste(outdir, str, ".interactions", sep="")

  # Get the variables
  group.vars <- unique(unlist(groups))

  op$check_interactions <- as.numeric(inter.flag)

  cnames     <- colnames(data) 
  allvars    <- NULL
  allvars    <- unique(c(response.var, vars0, group.vars, id.var))
  temp       <- !(allvars %in% cnames)
  if (any(temp)) {
    print("The following variables were not found in the input data frame")
    print(allvars[temp])
    stop()
  }

  if (op$check.data) {
    temp  <- samc.modifyData(data, response.var, id.var, vars0)
    vars0 <- temp$vars0
    if (!is.null(temp[["data", exact=TRUE]])) data <- temp$data
    rm(temp)
    gc() 
  } 
  
  # Get the column numbers
  response.col <- colNum(data, response.var)
  id.col       <- colNum(data, id.var)
  vars0.col    <- colNum(data, vars0)  
  if (group.flag) {
    groups.name <- removeWhiteSpace(names(groups))
    if (is.null(groups.name)) stop("group names must be specified")
    if (any(nchar(groups.name) < 1)) stop("group names must be specified")

    ngroups     <- length(groups)
    if (length(unique(groups.name)) != ngroups) stop("group names must be unique")
    groups.col  <- list()
    for (i in 1:ngroups) groups.col[[i]] <- colNum(data, unique(groups[[i]]))
    groups.num <- integer(ngroups)
  }
  
  if (response.var %in% vars0) stop("ERROR: vars0 cannot contain response.var")
  if (id.flag) {
    if (id.var %in% vars0) stop("ERROR: vars0 cannot contain id.var")
  }

  if (group.flag) {
    if (response.var %in% group.vars) stop("ERROR: groups cannot contain response.var")
    if (any(vars0 %in% group.vars)) stop("ERROR: vars0 cannot intersect with the groups")
    if (id.flag) {
      if (id.var %in% group.vars) stop("ERROR: groups cannot contain id.var")
    }
  }

  # Check interaction group names
  if (inter.flag) {
    if (!(all(unlist(interactions) %in% groups.name))) {
      stop("All groups specified in the interactions list must be defined in the groups list")
    }
  }

  # Vector of length ncol that gives the group each variable belongs to
  var_group  <- rep(0, ncol(data))

  # Create the groups file
  allcols <- c(response.col, vars0.col, id.col)
  if (group.flag) allcols <- c(allcols, unlist(groups.col))
  allcols <- unique(allcols)
  temp    <- (1:ncol(data))[-allcols]
  n       <- length(temp)
  x       <- NULL
  index   <- 1
  if (n) {
    for (i in 1:n) {
      jj <- temp[i]
      x  <- c(x, paste(index, "\t", jj, sep=""))
      var_group[jj] <- index
      index <- index + 1
    } 
  }
  if (group.flag) {
    for (i in 1:ngroups) {
      jj <- groups.col[[i]]
      if (any(as.logical(var_group[jj]))) stop("The same variable cannot belong to more than 1 group")
      x  <- c(x, paste(index, "\t", paste(jj, collapse=",", sep=""), sep=""))
      var_group[jj] <- index
      groups.num[i] <- index
      index <- index + 1
    }
  }
  write(x, file=group.file, ncolumns=1)

  # Create interactions file
  if (inter.flag) get_inter(interactions, groups.name, groups.num, inter.file)

  rm(groups.col, inter.col)
  gc()

  # Create the parameter file
  x <- "PARM VALUE"
  x <- c(x, paste("Data ", data.file, sep=""))
  x <- c(x, paste("groups ", group.file, sep=""))
  x <- c(x, paste("Data0_cols ", paste(vars0.col, collapse=",", sep=""), sep=""))
  x <- c(x, paste("response_col ", response.col, sep=""))
  if (inter.flag) x <- c(x, paste("interactions ", inter.file, sep=""))
  cnames <- names(op)
  for (i in 1:length(op)) {
    temp <- op[[i]]
    if (!is.null(temp)) x <- c(x, paste(cnames[i], " ", temp, sep=""))
  }
  write(x, file=parm.file, ncolumns=1)

  # Write out the data file
  write.table(data, file=data.file, sep=",", quote=FALSE, row.names=FALSE, col.names=TRUE)

  # Call the C function
  argc <- as.integer(1)
  argv <- c("", parm.file)
  ret  <- as.integer(-9)
  temp <- .C(func, argc, argv, ret=ret)
  if (temp$ret != 0) {
    if (temp$ret == -1) stop(paste("The option maxE=", op$maxE, " must be increased", sep=""))
    if (temp$ret == -2) stop("ERROR calling set_hist. Try changing the options lowE and/or maxE")
    stop(paste("ERROR: calling ", func, " C function", sep=""))
  }

  # Delete temporary files
  if (delete) {
    file.remove(data.file)
    file.remove(group.file)
    file.remove(parm.file)
    if (inter.flag) file.remove(inter.file)
  }

  # See if data was modified
  if (all(dim(data) == dim0)) data <- NULL
 
  for (v in outnames) {
    if (!file.exists(op[[v]])) op$v <- NULL
  }

  # Return list of output file names
  list(out_hist=op$out_hist, out1=op$out1, out2=op$out2, out_log=op$out_log,
       out_best=op$out_best, out_models=op$out_models, data=data)

} # END: mainfunc

# SAMC
BSR.samc <- function(data, outdir, response.var, id.var, vars0, groups=NULL, 
          interactions=NULL, op=NULL) {

  # data           Data frame containing only the varibles to be used in the analysis
  # outdir         Output directory to write output files and temporary files
  # response.var   Variable name of the response
  # vars0          Character vector of variables to always include in the models
  # groups         NULL or a list of character vectors defining the groups
  #                If NULL, then each variable is its own group.
  #                Example: groups=list(c("var1", "var2"), c("x1", "x5", "x34"))
  #                The default is NULL
  # op             List of options:
  #   stepscale         The default is 1
  #   total_iteration   The default is 500000
  #   warm              The default is 10000
  #   savestep          The default is 1
  #   lowE              The default is 0
  #   maxE              The default is 1000
  #   scale             The default is 1
  #   rho               The default is 1
  #   temperature       The default is 1
  #   maxvar            Maximum number of variables in a model. The default is 20.
  #   maxbeta           Maximum allowed value for abs(beta) in the cost function. The default is 1000.
  #   seed              The default seed is computed from the time. 
  #   gamma             The default is 1.0
  #   Model_maxn        Maximum number of models to save. The default is 100000.
  #   out.string        Character string to include in the output files. The default is "".
  #   delete            0 or 1 to delete temporary files. The default is 1.

  ret <- mainfunc(data, outdir, response.var, id.var, vars0, "samc", groups=groups, 
              interactions=interactions, op=op)
  ret

} # END: BSR.samc

# Resample
BSR.sample.model <- function(data, outdir, response.var, id.var, vars0, groups=NULL, 
              interactions=NULL, op=NULL) {

  if (is.null(op[["in_hist", exact=TRUE]])) {
    stop("The option in_hist must be specified") 
  }

  # See samc for input arguments
  ret <- mainfunc(data, outdir, response.var, id.var, vars0, "resample", groups=groups,
           interactions=interactions, op=op)
  ret

} # END: BSR.sample.model

