% --- Source file: man/samc.Rd ---
\name{BSR}
\alias{BSR}
\title{BSR}
\description{BSR is a wrapper function to combine the functions BSR.samc, BSR.sample.model and BSR.prediction}
\usage{
BSR(train.data, test.data, outdir, response.var, id.var, vars0, groups=NULL, 
     interactions=NULL, op=NULL)}

\arguments{
  \item{train.data}{Data frame containing only the variables to be used in the analysis. See details.}
  \item{test.data}{Data frame for testing that contains the same variables in the same order as \code{train.data}}.
  \item{outdir}{Directory where the output files will be written. If NULL, then the working directory will be used.}
  \item{response.var}{Variable name of the case-control response coded as 0=control, 1=case.}
  \item{id.var}{Variable name of the id variable or NULL if \code{train.data} and \code{test.data} do not contain
               an id variable.}
  \item{vars0}{Character vector of variables to always include in the models. Must be of length >= 1.}
  \item{groups}{NULL or a NAMED list of character vectors defining the groups. Each group of variables will always
                 appear together in a model. For example, if "var1" and "var2" are dummy variables for a 3 level
                 categorical variable, then \cr
                 \code{groups}=list(group1=c("var1", "var2")) will force variables var1 and var2
                 to always appear together in a model.
                 If NULL, then each variable is its own group. Any variable in the data frames other than 
                 \code{response.var} and \code{id.var}
                 which is not part of a group definition is treated as a group by itself. \cr
                Example: \code{groups}=list(grp1=c("var1", "var2"), grp2=c("x1", "x5", "x34")).
                 The default is NULL.}
  \item{interactions}{NULL or a list of character vectors of GROUP names that must be in the model
                    whenever the first GROUP name in the vector is in the model. 
                    The GROUP names must be a subset of the names in the \code{groups} list above, so that the 
                    \code{groups} argument must be specified to use this argument. This argument is used primarily 
                    for forcing the main effect variables of an interaction to be in the model whenever the
                    interaction variable(s) are in the model.
                    For instance, suppose that grp1=c("x1", "x2") is a group of dummy variables for a 3-level 
                    categorical variable,
                    and that grp2=c("z1", "z2", "z3") is a group of dummy variables for a 4-level categorical variable. 
                    The interaction of
                    the two categorical variables would be as group of 6 dummy variables \cr
                     grp3=c("x1z1", "x1z2", "x1z3", "x2z1", "x2z2", "x2z3").
                    If the \code{groups} argument is defined as \cr 
                    \code{groups}=list(grp1=c("x1", "x2"), grp2=c("z1", "z2", "z3"), grp3=c("x1z1", "x1z2", "x1z3", "x2z1",                    "x2z2", "x2z3")), then the \code{interactions} argument can be defined
                    as \cr
                  \code{interactions}=list(c("grp3", "grp1", "grp2")). Note that the first group in the vector is the group
                    for the interaction variables. 
                 The default is NULL.}
  \item{op}{List of options. See \code{details} for all possible options.}
}
\value{
  A list containing the prediction results (see \code{\link{BSR.prediction}}).
 
}
\details{
  The input data frames \code{train.data} and \code{test.data} will use ALL variables in the data frame as covariates 
    except \code{response.var} and \code{id.var}. All variables except for \code{id.var} must be numeric and continuous.
  Any interactions, dummy variables for categorical variables, and a column of ones for the intercept must already exist
  in the data frames before calling this function.
  
 \bold{Options list:} \cr

 Below are the names for the options list \code{op}. All names have default values
         if they are not specified. 
  \itemize{
  \item \code{out.string} String to be appended to output file names. The default is "".
  \item \code{train.op} Options list for training. See \code{\link{BSR.samc}}. 
    Note that the options \code{lowE} and \code{maxE} must be specified in this list.
    The default is NULL.
  \item \code{resample.op} Options list for resampling. See \code{\link{BSR.sample.model}}. 
        Note that the following options have no effect when calling the \code{BSR} function: 
     \code{stepscale}, \code{lowE}, \code{maxE}, \code{scale}, \code{rho}, \code{maxvar}, \code{maxbeta}, \code{gamma},
     \code{in_hist} and \code{in_models}.
   The default is NULL.
  \item \code{prediction.op} Options list for prediction. See \code{\link{BSR.prediction}}. The default is NULL.
  } % END: itemize

}
\seealso{\code{\link{BSR.sample.model}}, \code{\link{BSR.prediction}}, \code{\link{BSR.samc}}}
\author{Faming Liang, William Wheeler, Kai Yu}
\references{ 
Liang, F., Song, Q., and Yu, K. (2013).  Bayesian Subset Modeling for High Dimensional Generalized Linear Models.  J. Amer. Statist. Assoc., in press
 }
\examples{

  # Create data
  set.seed(123)
  n    <- 100
  m    <- 50
  y    <- rbinom(n, 1, 0.5)
  id   <- paste("sub_", 1:n, sep="")
  int  <- rep(1, n)
  z1   <- rbinom(n, 1, 0.5)
  z2   <- rbinom(n, 1, 0.5)
  z3   <- runif(n)
  z4   <- runif(n)
  data <- data.frame(y, id, int, z1, z2, z3, z4)

  train.data <- data[1:m, ]
  test.data  <- data[(m+1):n, ]
  outdir     <- "c:/temp/"
  vars0      <- "int"  # The intercept variable
  
  # Options for a small test run
  train.op      <- list(lowE=0, maxE=1000, total_iteration=100, warm=10)
  resample.op   <- list(total_iteration=100, warm=10)
  prediction.op <- list(n.burn=2) 
  op            <- list(train.op=train.op, resample.op=resample.op,
                        prediction.op=prediction.op, out.string="test")

  # Not run
  #ret <- BSR(train.data, test.data, outdir, "y", "id", vars0, groups=NULL,
  #       interactions=NULL,  op=op)

  
}