#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>

#define pi 3.14159265

#include "./samc.h"

/***********************************************************************************************
 NOTES:
  Input data set should contain an intercept column if an intercept is desired when fitting.
  The file of groups consists of 2 columns: 
    Column 1 is the group number labeled 1, 2, ..., N.
    Column 2 is the comma seperated list of column numbers.
  The file of interactions consists of 3 columns of group numbers: 
    Column 1 is the group number for the interaction.
    Column 2 is the group number for one of the main efects.
    Column 3 is the group number for the other main effect.

  The vector indx gives the original ordering for the new groups, so that indx[i] is the original
      group id of the ordered ith group.
  The vector ranks (inverse of indx) gives the rankings for the original group ids, so that ranks[i]
      is the ranking (new group id) for the original ith group.
**************************************************************************************************/

int mh_main(argc, argv)
int argc;
char *argv[];
{
  int stime;
  long ltime;
  int *x,*y,*truex,*besty,*indx,dim,tail,i,j,k,iter,grid;
  double **hist,*mar,*mary,w, logw, totalweight,delta,fvalue;
  double sum,ss,min,max,ave,un,var,*fx,*pw, *model;
  double deviance, skewness,z;
  FILE *ins=NULL, *ins2=NULL, *fpt_prob=NULL, *fpt_log=NULL, *fpt_hist=NULL;
  RET_COST *rcost;
  char *tempstr, sep;
  PARMS *parms;
  int nc, nr, *ivec, header, **tempX, interFlag, grp, **tmpMat, checkInt, num_grp, k0, histFlag;
  double rho, scale, lowE, maxE, **DataX, *DataY, *refden, *cor, *fsave, bestvalue, accept_loc, total_loc;
  int stepscale, savestep, warm, total_iteration, DataNum, DataP, TotalNcol, Data0X_n, maxvar, *Data0X;
  int **GroupX, *GroupN, Sum_GroupN, **IntX, IntX_flag, sze, *bestx, **modsave, *ranks, *IntX_N, *ivec2;

  /* Get the parameters */
  parms = (PARMS *) malloc(sizeof(PARMS));
  CHECK_MEM(parms);

  /* Initialize some parms before parm file is read */
  parms->Model_maxn = MODEL_MAXN;
  parms->seed       = -1;
  parms->maxbeta    = MAX_BETA;

  /* Read the parameter file */
  read_parms(parms, argv[1]);
  total_iteration = parms->total_iteration;
  warm            = parms->warm;
  savestep        = parms->savestep;
  lowE            = parms->lowE;
  maxE            = parms->maxE;
  scale           = parms->scale;
  rho             = parms->rho;
  stepscale       = parms->stepscale;
  checkInt        = parms->check_interactions;

  /* Set up the models structure */
  set_Models(argv[1], parms);
  
  /* Allocate memory for temporary character strings */
  tempstr = (char *) malloc(MAXFILELEN*sizeof(char));
  CHECK_MEM(tempstr);

  /* Get the name of the input file */
  get_option(argv[1], "Data", tempstr, 1);

  /* Get the file delimiter */
  sep = getDelimiter(tempstr, BUFFERSIZE);

  /* Get the number of rows and columns */
  nr = getNrows(tempstr, BUFFERSIZE);
  nc = getNcols(tempstr, sep, BUFFERSIZE);

  /* Determine if the file has a header of variable names */
  header = getHeader(tempstr, sep, nc, BUFFERSIZE);
  if (header) nr--;
  DataNum = nr;
  parms->DataNum = DataNum;
  TotalNcol = nc;
  parms->TotalNcol = TotalNcol;

  printf("Nrows=%d, Ncols=%d, header=%d\n", nr, nc, header);

  /* Read the data into a matrix */
  DataX = dmatrix(1, nr, 1, nc); 
  parms->DataX = DataX;
  read_data(tempstr, sep, DataX, nr, nc, BUFFERSIZE, header);

  print_dMat(DataX, 1, 10, 1, MIN(10, nc), "Subset of data file:");

  /* Store the response */
  DataY = dvector(1, DataNum);
  parms->DataY = DataY;
  k = parms->response_col;
  for (i=1; i<=DataNum; i++) DataY[i] = DataX[i][k];

  /* Get the columns to be always included */
  get_option(argv[1], "Data0_cols", tempstr, 1);
  ivec = ivector(0, nc+1);
  k = getIntsFromStr(tempstr, 10, nc+1, ivec);

  /* Get the number of cols to always include */
  for (i=1, k=0; i<=nc; i++) k += ivec[i];
  Data0X_n = k;
  parms->Data0X_n = Data0X_n;

  /* Put the column number always to include in Data0X vector */
  if (Data0X_n) {
    Data0X = ivector(1, Data0X_n);
    for (i=1, k=0; i<=nc; i++) {
      if (ivec[i]) {
        k++;
        Data0X[k] = i;    
      }
    }
  } else {
    Data0X = NULL;
  }
  parms->Data0X = Data0X;

  /* Define the groups. Groups must be defined */
  ivec[parms->response_col] = 1;
  get_option(argv[1], "groups", tempstr, 1);

  tempX = imatrix(1, nc, 1, nc);
    
  /* Read in the file of groups and set up GroupX and GroupN */
  DataP  = read_groups(tempstr, ivec, nc, tempX);
  parms->DataP = DataP;
  maxvar = MIN(parms->maxvar, DataP);

  GroupX = imatrix(1, DataP, 1, DataP);
  GroupN = ivector(1, DataP);
  Sum_GroupN = 0;
  for (i=1, grp=0; i<=nc; i++) {
    /* Get the number of columns in this group */
    for (j=1, k=0; j<=nc; j++) k += tempX[i][j];
    if (k) {
      grp++;
      GroupN[grp] = k;
      Sum_GroupN += k;
      for (j=1, k=0; j<=nc; j++) {
        if (tempX[i][j]) {
          k++;
          GroupX[grp][k] = j;
        }
      }
    }
  }
  parms->GroupX = GroupX;
  parms->GroupN = GroupN;
  parms->Sum_GroupN = Sum_GroupN;

  if (tempX) free_imatrix(tempX,1,nc,1,nc);
  free_ivector(ivec, 0, nc+1);

  /* Set up the interaction matrix, not using this feature at this moment */
  interFlag = 1 - get_option(argv[1], "interactions", tempstr, 0);

  if (interFlag) {
    /* The interaction data file can have a different number of columns for each row.
       Column 1 is the interaction group number, followed by the other group numbers that must be in the model.
    */
    IntX_N = ivector(1, DataP);
    IntX = read_interactions(tempstr, DataP, IntX_N);
    IntX_flag = 1;
  } else {
    IntX_flag = 0;
    IntX = NULL;
    IntX_N = NULL;
  }
  parms->IntX_flag = IntX_flag;
  parms->IntX = IntX;
  parms->IntX_N = IntX_N;

  /* Open the output files */
  i = get_option(argv[1], "out1", tempstr, 0);
  if (!i) ins = open_file(tempstr, "w");
  i = get_option(argv[1], "out2", tempstr, 0);
  if (!i) ins2 = open_file(tempstr, "w");

  i = get_option(argv[1], "out_prob", tempstr, 0);
  if (!i) fpt_prob = open_file(tempstr, "w");

  i = get_option(argv[1], "out_log", tempstr, 0);
  if (!i) fpt_log = open_file(tempstr, "w");
  i = get_option(argv[1], "out_hist", tempstr, 0);
  if (!i) fpt_hist = open_file(tempstr, "w");

  free(tempstr);

  rcost = cost_init(DataNum, DataP, TotalNcol);
  /*rcost_best = cost_init(DataNum, DataP, TotalNcol); */

  /* initialize the random number generator */
  if (parms->seed > -1) {
    srand(parms->seed);
    if (fpt_log) {
      fprintf(fpt_log, "random seed=%d \n", parms->seed);
      fflush(fpt_log);
    }
  } else {
    ltime=time(NULL);
    stime=(unsigned int)ltime/2;
    srand(stime);
    if (fpt_log) {
      fprintf(fpt_log, "random seed=%d \n", stime);
      fflush(fpt_log);
    }
  }

  /* Set up the hist matrix. hist matrix starts at 0 */
  grid = ceil((maxE-lowE)*scale);
  hist = set_hist(argv[1], &grid, &histFlag, parms);
  sze  = grid;
  parms->sze = sze;

  refden=dvector(0,grid);
  parms->refden = refden;

  x=ivector(1,DataP);
  y=ivector(1,DataP);
  truex=ivector(1,DataP);
  bestx=ivector(1,DataP);
  parms->bestx = bestx;
  besty=ivector(1,DataP);
  /*DataZ=dmatrix(1,DataNum,1,DataP);*/
  mar=dvector(1,DataP);
  mary=dvector(1,DataP);
  pw=dvector(1,DataP);
  fx=dvector(1,DataP);
  indx=ivector(1,DataP);
  parms->indx = indx;
  cor=dvector(1,DataP);
  parms->cor = cor;
 
  /* the 0 position is for the model using only offset */
  modsave=imatrix(0,maxvar,1,DataP);
  parms->modsave = modsave;
  fsave=dvector(0,maxvar); 
  parms->fsave = fsave;
  model = dvector(0, maxvar);
  for (i=0; i<=maxvar; i++) model[i] = 0.0;

  /*PenaltyC=log(1.0*DataNum)+2.0*gamma*log(1.0*DataP);*/
  /*  PenaltyC=log(1.0*DataNum)+2.0*gamma*log(1.0*DataP); */

  ranks = ivector(1, DataP);
  for (i=1; i<=DataP; i++) {
    ranks[i] = i;
    indx[i]  = i;
  }
  parms->ranks = ranks;

  /* start from here */
 
  /* rank individual variable */
  if (DEBUG) printf("Begin: Rank variables\n");
  min=1.0e+100; 
  max=-1.0e+100;
  for(sum=var=0.0, i=1; i<=DataP; i++){
     for(j=1; j<=DataP; j++) x[j]=0;
     x[i]=1; /* changed */
     cost(x, rcost, 0, parms);  /* Do not check interactions here */
     fx[i] = rcost->value;

     if(fx[i]<1.0e+100){ sum+=fx[i]; var+=fx[i]*fx[i]; }
     if(fx[i]<min) min=fx[i];
       else if(fx[i]>max) max=fx[i];

    }


  ave=sum/DataP; 
  var=(var - DataP*ave*ave)/(DataP-1);
  for(i=1; i<=DataP; i++)
    {
       pw[i]=-1.0*exp(-(fx[i]-min)/(5.0*sqrt(var))-0.1);
       if(pw[i]> -0.01) pw[i]=-0.01; 
    }

 
  if (DEBUG) printf("END: Rank variables\n");  

  /* indx gives the grp id for the ith ranked groups*/
  indexx(DataP, pw, indx);

  /* Reorder x and change the values in GroupX, GroupN, and intX */
  tempX  = imatrix(1, DataP, 1, DataP);
  tmpMat = imatrix(1, DataP, 1, DataP);
  ivec   = ivector(1, DataP);
  ivec2  = ivector(1, DataP);
  for(i=1; i<=DataP; i++){
    k = indx[i];
    ranks[k] = i;
    ivec[i] = GroupN[k];
    for (j=1; j<=DataP; j++) tempX[i][j] = GroupX[k][j];
    if (IntX_flag) {
      for (j=1; j<=IntX_N[k]; j++) tmpMat[i][j] = IntX[k][j-1];
      ivec2[i] = IntX_N[k];
    }
  }
  for(i=1; i<=DataP; i++){
    GroupN[i] = ivec[i];
    for (j=1; j<=DataP; j++) GroupX[i][j] = tempX[i][j];
    if (IntX_flag) {
      k = ivec2[i];
      /* Free memory, then re-allocate */
      if (IntX[i]) free(IntX[i]);
      IntX[i] = (int *) malloc(k*sizeof(int));
      CHECK_MEM(IntX[i]);
      for (j=1; j<=k; j++) IntX[i][j-1] = tmpMat[i][j];
      IntX_N[i] = k;
    }
  }
  free_ivector(ivec, 1, DataP);
  free_ivector(ivec2, 1, DataP);
  free_imatrix(tempX, 1, DataP, 1, DataP);
  free_imatrix(tmpMat, 1, DataP, 1, DataP);

  /* cor is the ordered pw */
  for(j=1; j<=DataP; j++){
      cor[j]=-1.0*pw[indx[j]];
  }

  /* Initialization */
  for(sum=0.0, ss=1.0, i=0; i<=sze; i++){ ss+=i; refden[i]=1.0; sum+=refden[i]; }
  for(i=0; i<=sze; i++) refden[i]/=sum;
  
  /* don't initizlize it
  for(i=0; i<=sze; i++){
     hist[i][1]=lowE+i*1.0/scale; 
     hist[i][2]=0.0;
     hist[i][3]=0.0;
  }
  */


  /*so the last row is maxE+1.0/scale */    
  /* ---hist[0]---hist[1]---hist[2]---max---max+1 */

    /* initialization of samx */
  for(sum=0.0, i=1; i<=DataP; i++){ x[i]=0; y[i]=i; sum+=cor[i]; mar[i]=cor[i]; }

  /* randomly select a inital x to start, can change so start from the same */  
  for(i=1; i<=2; i++){ 
        un=rand()*1.0/RAND_MAX;
        ss=mar[1]/sum; k=1;
        while(un>ss && k<DataP-i+1){
              k++;
              ss+=mar[k]/sum;
         } 
         x[y[k]]=1; sum-=mar[k];
         for(j=k; j<=DataP-i; j++){ mar[j]=mar[j+1]; y[j]=y[j+1]; }
  } 
  

  cost(x, rcost, checkInt, parms);
  fvalue = rcost->value;
 
  bestvalue=fvalue;
  parms->bestvalue = bestvalue;

  /* dim should be the number of columns in x */
  for(dim=0, i=1; i<=DataP; i++)
  { 
     bestx[i]=x[i]; 
  }

  /*deviance=2.0*fvalue-dim*PenaltyC; */
  
  deviance = rcost->deviance;

  dim = rcost->dim;
  /* the degree of freedom need to be changed */ 
  un=gammpln(0.5*(DataNum-dim),0.5*deviance,&tail);
  z=tail*inverse_normal_cdf_log(un);
  
    
  accept_loc=total_loc=0.1; totalweight=0.0; skewness=0.0;
  parms->accept_loc = accept_loc;
  parms->total_loc  = total_loc;

   /**********************************************************************/
   for(i=1; i<=DataP; i++) mar[i]=0.0; 
   for(i=0; i<=maxvar; i++) fsave[i]=1.0e+200;

   for(iter=1; iter<=warm+total_iteration; iter++){
      
       if (DEBUG) printf("****** ITER = %d ******\n", iter);
       if(iter<=stepscale) delta=rho;
       else delta=rho*stepscale/iter;
    
       if (DEBUG) printf("Begin metmove\n");
       Metropolis_ini(x,&fvalue, rcost, checkInt, parms);
        
       if (DEBUG) printf("End metmove\n");

       deviance=rcost->deviance;
       dim = rcost->dim;
       num_grp = rcost->num_grp;  
    
       /*below, the y vector is the oringal order */
       indexx_convert_integer(DataP,indx,x,y);
       if (ins) {
	 fprintf(ins, " %d %d %g %g \n", dim,num_grp, deviance,
		 fvalue); 
	 fflush(ins);
       }

        
       if(iter%100==0) printf(" iter=%d bestv=%g  fval=%g \n",iter,parms->bestvalue,fvalue);
       

   } /* END: for(iter=1; iter<=warm+total_iteration; iter++) */
  
   if (ins) fclose(ins); 
   if (fpt_prob) fclose(fpt_prob);
 
   
    /* Write out models */
    write_Models(argv[1], parms);

  free_dmatrix(hist,0,grid,1,4);
  free_dvector(refden,0,grid);
  free_ivector(x,1,DataP);
  free_ivector(y,1,DataP);
  free_ivector(truex,1,DataP);
  free_ivector(bestx,1,DataP);
  free_ivector(besty,1,DataP);
  if (DataX != NULL) free_dmatrix(DataX,1,DataNum,1,TotalNcol);
  if (Data0X != NULL) free_ivector(Data0X,1,Data0X_n);
  /*free_dmatrix(DataZ,1,DataNum,1,DataP);*/
  free_dvector(DataY,1,DataNum);
  free_dvector(mar,1,DataP);
  free_dvector(mary,1,DataP);
  free_dvector(pw,1,DataP);
  free_dvector(fx,1,DataP);
  free_ivector(indx,1,DataP);
  free_dvector(cor,1,DataP);
  free_imatrix(modsave,0,maxvar,1,DataP);
  free_dvector(fsave,0,maxvar);
  free_dvector(model,0,maxvar);
  free_ivector(GroupN,1,DataP);
  free_imatrix(GroupX,1,DataP,1,DataP);
  free_ivector(ranks, 1, DataP);

  /* for the purpose of free the object */
  rcost->num_grp = DataP;
  rcost->dim     = TotalNcol;
  cost_free(rcost, DataNum);

  if (parms->IntX_N) free_ivector(parms->IntX_N, 1, DataP);
  if (parms->IntX) matrix_free((void **) parms->IntX, DataP+1);

  models_free(parms);
  free(parms);

  return(0);

} /* END: mh_main */

/* Function that R will call */
void mh(pargc, argv, ret)
int *pargc, *ret;
char **argv;
{
  *ret = 1;
  *ret = mh_main(*pargc, argv);

  return;

} /* END: samc */




